from pprint import pprint
import sys
import os
import argparse
import multiprocessing

from smear.file_identifier import detect_mc_type, McFileTypes
from smear.uiterm import markup_print as mprint
from smear.smearing import print_detectors, detectors

help_text = '''
Simple command line interface to eic-smear and jleic-smear

usage: 
  ejana_smear <flags> [input file]   # smears the input file
  ejana_smear -l                     # list of available detectors

examples:
  ejana_smear beagle_eD.txt                     # smear the file with handbook detector
  ejana_smear beagle_eD.txt -o my_file.root     # smears the file and set output file name  
  ejana_smear -d beast -n 10000 beagle_eD.txt   # uses beast detector, smears only 10k events
   
main flags:    
  -l / --list           Show all detectors and versions    
  -d / --detector       Detector parametrisation. Handbook is default. 
                        Run with -l/--list flag to see all options
  -n / --nevents        Number of events to process
  -s / --nskip          Number of events to skip    
  -o / --output         Output file name    
  -j / --threads        Number of threads. Set 'auto' for max. Default is 1  
  -h / --help           Show help

advanced flags:
  -t / --input-type     Input file type. Select from: beagle, hepmc2, lund, eic_pythia6, g4e, auto
  -a / --analysis       Comma separated additional analysis list
  --explain             Explain all configurations, but don't run smearing

number of events:
  One can set a number of events by -n(--nevents) and -s(--nskip) flags. Flag -n also support ranges:
    -n 100        : process 100 events
    -s 100 -n 50  : skip first 100 events, and process 50 events
    -s 100        : skip first 100 events and process the rest

known file formats:
  Applies to -t / --input-type
    beagle, hepmc2, lund, eic_pythia6, g4e, auto
'''

type_to_plugin = {
    'beagle': 'beagle_reader',
    'hepmc2': 'hepmc_reader',
    'eic_pythia6': 'lund_reader',
    'lund': 'lund_reader',
    'g4e': 'g4e_reaer'
}

def _get_jana_args(params):
    """Get Jana constructor arguments from general paramters"""
    jana_args = {}
    if params.nevents:
        jana_args['nevents'] = params.nevents

    if params.nskip:
        jana_args['nskip'] = params.nskip

    jana_args['output'] = params.output
    jana_args['nthreads'] = params.threads
    return jana_args


def run_smearing(params):
    from pyjano.jana import Jana

    jana_args = _get_jana_args(params)

    jana = Jana(**jana_args)
    jana.plugin(type_to_plugin[params.input_type])
    jana.plugin('event_writer')
    jana.plugin('eic_smear', detector=params.detector)
    jana.source(params.input_file)

    if params.explain:
        mprint("<b><blue>run command  : </blue></b> ejana " + jana.get_run_command())
    else:
        jana.run()


def makeup_output_file_name(input_file):
    file_name = os.path.basename(input_file)
    base, ext = os.path.splitext(file_name)
    return base + ".root"


def process_arguments(args):

    # First, we parse arguments using argparse
    params = parse_arguments(args)

    # Result will be argparse.Namespace or int with suggested return code
    if isinstance(params, int):
        return params

    # print input file
    mprint("<b><blue>input  file  : </blue></b>" + params.input_file)
    # Test the file exists
    if not os.path.isfile(params.input_file):
        mprint("<red>(!) Error (!)</red> File does not exist or there is no permission to read it. "
               "File name is '{}' \n", params.input_file)
        return 1

    # Validate the detector
    if params.detector:
        if params.detector not in detectors.keys():
            mprint("<red>(!) Error (!)</red> Detector '{}' is not a known detector name/version. "
                   "Please run with -l flag to see the list of all available detectors".format(params.detector))
            return 1

    # determine file type
    if not params.input_type or params.input_type == 'auto':
        params.input_type = detect_mc_type(params.input_file).value
        if params.input_type == 'unknown':
            mprint("<red>(!) Error (!)</red> Cannot automatically determine file type. "
                   "Use -t/--input-type flag to set file type manually. Known types are: " + McFileTypes.list())
            return 1

        mprint("<b><blue>input  type  : </blue></b>" + params.input_type + " <green>(auto)</green>")
    else:
        mprint("<b><blue>input  type  : </blue></b>" + params.input_type)

    # makeup output file name
    if not params.output:
        params.output = makeup_output_file_name(params.input_file)
        mprint("<b><blue>output file  : </blue></b>" + params.output + " <green>(auto)</green>")
    else:
        mprint("<b><blue>output file  : </blue></b>" + params.output)

    # Number of events to skip
    if params.nskip:
        mprint("<b><blue>skip events  : </blue></b> " + str(params.nskip))
    else:
        mprint("<b><blue>skip events  : </blue></b>don't skip")

    # Number of events to process
    if params.nevents:
        mprint("<b><blue>proc events  : </blue></b>" + str(params.nevents))
    else:
        mprint("<b><blue>proc events  : </blue></b>until the end of file")

    # Number of threads
    available_threads = multiprocessing.cpu_count()
    if not params.threads:
        params.threads = 1
        mprint("<b><blue>num threads  : </blue></b>{} out of {} <green>(default)</green>",
               params.threads, available_threads)
    else:
        if params.threads == 'auto':
            params.threads = available_threads
            mprint("<b><blue>num threads  : </blue></b>{} <green>(auto)</green>", params.threads)
        else:
            params.threads = int(params.threads)
            mprint("<b><blue>num threads  : </blue></b>{}", params.threads)

    # detector
    if not params.detector:
        params.detector = 'matrix'
        mprint("<b><blue>detector     : </blue></b>" + params.detector + " latest <green>(default)</green>")
    else:
        mprint("<b><blue>detector     : </blue></b>" + params.detector)

    return params


def parse_arguments(args):
    """
    Parses program arguments


    :param args: sys.argv or test arguments
    :return: Returns either integer with the exist code or parse result
    """


    # noinspection PyTypeChecker
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False)

    parser.add_argument('input_files', nargs='*', default=None)

    # "List detectors"
    parser.add_argument("-l", "--list", action="store_true", help=argparse.SUPPRESS)    # List all available detectors
    parser.add_argument("-a", "--analysis", type=str, help=argparse.SUPPRESS)           # Comma separated additional analysis list
    parser.add_argument("-d", "--detector", type=str, help=argparse.SUPPRESS)           # Detector parametrisation. Handbook is default. Run -l/--list flag to see all options
    parser.add_argument("-t", "--input-type", default='auto', help=argparse.SUPPRESS)   # Input file type. Select from: beagle, hepmc2, lund, g4e
    parser.add_argument("-o", "--output", help=argparse.SUPPRESS)                       # Output file name
    parser.add_argument("-n", "--nevents", type=int, help=argparse.SUPPRESS)            # Number of events to process
    parser.add_argument("-s", "--nskip", type=int, help=argparse.SUPPRESS)              # Number of events to skip. See details for ranges
    parser.add_argument("-j", "--threads", help=argparse.SUPPRESS)                      # Number of threads
    parser.add_argument("-h", "--help", action="store_true", dest='show_help')          # Show help
    parser.add_argument("--explain", action="store_true")                               # Just explain the setup but don't run enything

    result = parser.parse_args(args)

    if len(sys.argv) == 0:
        print("(!) Provide at least one file to process. Run with --help flag")
        print(help_text)
        return 1

    if result.show_help:
        print(help_text)
        return 0

    if result.list:
        print_detectors()
        return 0

    # Validate input files
    if not result.input_files:
        mprint("<red>(!) Error (!)</red> Please provide at least one input file")
        print(help_text)
        return 1

    if len(result.input_files) > 1:
        mprint("<red>(!) Error (!)</red> Unfortunately, only a single file can be processed at a time. "
               "It will be updated in further versions")
        return 1

    result.input_file = result.input_files[0]

    return result


def smear_cli(args=None):
    """Run smearing according to user arguments"""
    # Print the setup and auto-identify missing params
    program_params = process_arguments(args if args is not None else sys.argv[1:])

    # Result will be argparse.Namespace or int with suggested return code
    if isinstance(program_params, int):
        exit(program_params)
    run_smearing(program_params)

    return program_params


if __name__ == "__main__":
    pp = smear_cli()
    pprint(pp)





_ansi_colors = {
    "black": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "white": 37,
    "reset": 39,
    "bright_black": 90,
    "bright_red": 91,
    "bright_green": 92,
    "bright_yellow": 93,
    "bright_blue": 94,
    "bright_magenta": 95,
    "bright_cyan": 96,
    "bright_white": 97,
}
_ansi_reset_all = "\033[0m"



# A list of known text names of colors
known_colors = ["black",        # (might be a gray)
                "red",
                "green",
                "yellow",       # (might be an orange)
                "blue",
                "magenta",
                "cyan",
                "white",        # (might be light gray)
                "bright_black",
                "bright_red",
                "bright_green",
                "bright_yellow",
                "bright_blue",
                "bright_magenta",
                "bright_cyan",
                "bright_white",
                "reset"]


def style(
    text,
    fg=None,
    bg=None,
    bold=None,
    dim=None,
    underline=None,
    blink=None,
    reverse=None,
    reset=True,
):
    """Styles a text with ANSI styles and returns the new string.  By
    default the styling is self contained which means that at the end
    of the string a reset code is issued.  This can be prevented by
    passing ``reset=False``.
    Examples::
        click.echo(click.style('Hello World!', fg='green'))
        click.echo(click.style('ATTENTION!', blink=True))
        click.echo(click.style('Some things', reverse=True, fg='cyan'))
    Supported color names:
    * ``black`` (might be a gray)
    * ``red``
    * ``green``
    * ``yellow`` (might be an orange)
    * ``blue``
    * ``magenta``
    * ``cyan``
    * ``white`` (might be light gray)
    * ``bright_black``
    * ``bright_red``
    * ``bright_green``
    * ``bright_yellow``
    * ``bright_blue``
    * ``bright_magenta``
    * ``bright_cyan``
    * ``bright_white``
    * ``reset`` (reset the color code only)
    .. versionadded:: 2.0
    .. versionadded:: 7.0
       Added support for bright colors.
    :param text: the string to style with ansi codes.
    :param fg: if provided this will become the foreground color.
    :param bg: if provided this will become the background color.
    :param bold: if provided this will enable or disable bold mode.
    :param dim: if provided this will enable or disable dim mode.  This is
                badly supported.
    :param underline: if provided this will enable or disable underline.
    :param blink: if provided this will enable or disable blinking.
    :param reverse: if provided this will enable or disable inverse
                    rendering (foreground becomes background and the
                    other way round).
    :param reset: by default a reset-all code is added at the end of the
                  string which means that styles do not carry over.  This
                  can be disabled to compose styles.
    """
    bits = []
    if fg:
        try:
            bits.append(f"\033[{_ansi_colors[fg]}m")
        except KeyError:
            raise TypeError(f"Unknown color {fg!r}")
    if bg:
        try:
            bits.append(f"\033[{_ansi_colors[bg] + 10}m")
        except KeyError:
            raise TypeError(f"Unknown color {bg!r}")
    if bold is not None:
        bits.append(f"\033[{1 if bold else 22}m")
    if dim is not None:
        bits.append(f"\033[{2 if dim else 22}m")
    if underline is not None:
        bits.append(f"\033[{4 if underline else 24}m")
    if blink is not None:
        bits.append(f"\033[{5 if blink else 25}m")
    if reverse is not None:
        bits.append(f"\033[{7 if reverse else 27}m")
    bits.append(text)
    if reset:
        bits.append(_ansi_reset_all)
    return "".join(bits)


# noinspection PyBroadException
try:
    # noinspection PyUnresolvedReferences
    from HTMLParser import HTMLParser
except:
    from html.parser import HTMLParser


class ColorMarkupClickPrinter(HTMLParser, object):

    def error(self, message):
        pass

    is_bold = False       # the next printed text will be bold
    is_underline = False  # the next printed text will be underlined
    color = ""            # color of the next printed text

    def handle_starttag(self, tag, attrs):

        # make in case independent
        tag = tag.lower()

        # do we know this tag?
        if tag == 'b':
            self.is_bold = True
        elif tag == 'u':
            self.is_underline = True
        elif tag in known_colors:
            self.color = tag
        else:
            print(style('<{}>'.format(tag), fg=self.color, bold=self.is_bold, underline=self.is_underline), end="")


    def handle_endtag(self, tag):
        # make in case independent
        tag = tag.lower()

        # do we know this tag?
        if tag == 'b':
            self.is_bold = False
        elif tag == 'u':
            self.is_underline = False
        elif tag in known_colors:
            self.color = "reset"
        else:
            print(style('<{}>'.format(tag), fg=self.color, bold=self.is_bold, underline=self.is_underline), end="")

    def handle_data(self, data):
        print(style(data, fg=self.color, bold=self.is_bold, underline=self.is_underline), end="")


def markup_print(text, *args, **kwargs):
    """markdown print
    <red> </red>
    <green> <green>
    <blue><blue>
    """
    if args or kwargs:
        text = text.format(*args, **kwargs)
    parser = ColorMarkupClickPrinter()
    parser.feed(text)
    print("")     # new line in the end


if __name__ == "__main__":
    markup_print("One two <b> df <red> my red red red </red> </b>")


import enum
from smear.uiterm import markup_print


class SmearingEngines:
    EIC_SMEAR = 'eic_smear'
    JLEIC_SMEAR = 'jleic_smear'


class DetectorDescription:
    def __init__(self, version='', description='', flag='', engine='eic_smear', link=''):
        self.version = version
        self.description = description
        self.engine = engine
        self.link = link


detectors = {
    "matrix": DetectorDescription(
            version='v0.1',
            description="Yellow Reports Detector Working Goroup detector based on the Detector Matrix (alias to matrix-0.1)",
            link='https://physdiv.jlab.org/cgi-bin/DetectorMatrix/view',
            engine=str(SmearingEngines.EIC_SMEAR)
        ),

    "matrix-0.1": DetectorDescription(
            version='v0.1',
            description="Strict handbook table (simple smear engine by YF)",
            link='eic_smear/ESDetectorHandBook_v1_0_4.cc',
            engine=str(SmearingEngines.JLEIC_SMEAR)
        ),
    "smatrix": DetectorDescription(
        version='v0.0',
        description="Strict YP DWG Matirx table (simple smear engine by YF)",
        link='eic_smear/ESDetectorHandBook_v1_0_4.cc',
        engine=str(SmearingEngines.JLEIC_SMEAR)
    ),
    "handbook": DetectorDescription(
        version='v1.2',
        description="Based on HANDBOOK v1.2 (Feb 20, 2020) with adaptations (alias to handbook-1.2)",
        link='http://www.eicug.org/web/content/detector-rd',
        engine=str(SmearingEngines.EIC_SMEAR)
    ),

    "handbook-1.2": DetectorDescription(
        version='v1.2',
        description="Based on HANDBOOK v1.2 (Feb 20, 2020) with adaptations",
        link='http://www.eicug.org/web/content/detector-rd',
        engine=str(SmearingEngines.EIC_SMEAR)
    ),

    "beast": DetectorDescription(
        version='v0.1',
        description="BeAST (alias to beast 0.1)",
        link='ESDetectorBeAST_v1_0_4.cc',
        engine=str(SmearingEngines.EIC_SMEAR)
    ),

    "beast-0.1": DetectorDescription(
        version='v0.1',
        description="BeAST detector ",
        link='https://arxiv.org/abs/1708.01527',
        engine=str(SmearingEngines.EIC_SMEAR)
    ),
    "beast-0.0": DetectorDescription(
        version='v0.0',
        description="BeAST",
        link='https://arxiv.org/abs/1708.01527',
        engine=str(SmearingEngines.EIC_SMEAR)
    ),
    "jleic": DetectorDescription(
        version='0.2',
        description="JLEIC detector (simple smear engine by YF)",
        link='',
        engine=str(SmearingEngines.JLEIC_SMEAR)
    ),
    "jleic-0.2": DetectorDescription(
        version='0.2',
        description="JLEIC (simple smear engine)",
        link='',
        engine=str(SmearingEngines.JLEIC_SMEAR)
    ),
    "jleic-0.1": DetectorDescription(
        version='0.1',
        description="JLEIC (simple smear engine)",
        link='',
        engine=str(SmearingEngines.JLEIC_SMEAR)
    ),
    "ephenix": DetectorDescription(
        version='v0.0',
        description="ePHENIX detector (Example from 2014)",
        link='ePHENIX',
        engine=str(SmearingEngines.EIC_SMEAR)
    ),
    "ephenix_nms": DetectorDescription(
        version='v0.0',
        description="ePHENIX detector (Example from 2014). No multiple scattering",
        link='https://arxiv.org/abs/1708.01527',
        engine=str(SmearingEngines.EIC_SMEAR)
    ),
    "estar": DetectorDescription(
        version='v0.0',
        description="An example implementation from 2014",
        link='https://wiki.bnl.gov/conferences/index.php/January_2014',
        engine=str(SmearingEngines.EIC_SMEAR)
    ),
    "star": DetectorDescription(
        version='v0.0',
        description="STAR parametrization example",
        link='https://wiki.bnl.gov/conferences/index.php/January_2014',
        engine=str(SmearingEngines.EIC_SMEAR)
    ),
    "zeus": DetectorDescription(
        version='v0.0',
        description="Example of ZEUS detector",
        link='ESDetectorZeus_v1_0_0.cc',
        engine=str(SmearingEngines.EIC_SMEAR)
    ),
    "perfect": DetectorDescription(
        version='0.0',
        description="Perfect detection and PID in |η| < 15",
        link='',
        engine=str(SmearingEngines.JLEIC_SMEAR)
    ),
}


def print_detectors():
    """ Pretty prints detectors list"""
    for name, detector in detectors.items():
        markup_print("\n<blue>detector</blue>: {}".format(name))
        markup_print(" <b>version</b> : {}".format(detector.version))
        markup_print(" <b>descr.</b>  : {}".format(detector.description))
        markup_print(" <b>link</b>    : {}".format(detector.link))
        markup_print(" <b>engine</b>  : {}".format(detector.engine))


if __name__ == '__main__':
    print_detectors()



import enum
import os


class McFileTypes (enum.Enum):
    BEAGLE = "beagle"
    HEPMC2 = "hepmc2"
    EIC_PYTHIA6 = "eic_pythia6"
    LUND = "lund"
    G4E = "g4e"
    UNKNOWN = "unknown"

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


def detect_mc_type(file_name):
    """Automatically detects file type. Returns one of McFileTypes"""
    filename, extension = os.path.splitext(file_name)

    if 'root' in extension:
        raise ValueError("Auto identification of .root files is not yet implemented. "
                         "Please use --input-format flag. Example '--imput-format=g4e'")
    return detect_text_mc_type(file_name)


def detect_text_mc_type(file_name):
    """Automatically detects text file types. Returns one of McFileTypes"""

    with open(file_name, 'r') as f:
        line = ""
        try:
            while not line:                 # Just to fake a lot of readlines and hit the end
                line = next(f).replace('\n', '')
        except StopIteration:
            return McFileTypes.UNKNOWN
        if "BEAGLE EVENT FILE" in line:
            return McFileTypes.BEAGLE
        elif "HepMC" in line:
            return McFileTypes.HEPMC2
        elif "PYTHIA EVENT FILE" in line:
            return McFileTypes.EIC_PYTHIA6
        elif len(line.split()) == 10:
            return McFileTypes.PYTHIA_LUND
    return McFileTypes.UNKNOWN